use num_bigint::{BigUint, ToBigUint};
use num_traits::One;

#[cfg(not(target_arch = "wasm32"))]
use rand::thread_rng;

#[cfg(not(target_arch = "wasm32"))]
use num_bigint::RandBigInt;

/// Sloth is a verifiable delay function (VDF) that is designed to enforce
/// a predetermined amount of sequential work.

pub struct Sloth {
    /// The modulus `p` is a large prime number.
    pub p: BigUint,
}

impl Sloth {
    /// Create a new Sloth VDF with the given modulus `p` and iteration count `t`.
    #[cfg(not(target_arch = "wasm32"))]
    pub fn new(p: BigUint) -> Self {
        Sloth { p }
    }

    /// Encode a message `x` using the Sloth VDF.
    #[cfg(not(target_arch = "wasm32"))]
    pub fn encode(&self, x: &BigUint, t: u32) -> BigUint {
        let mut y = x.clone();

        // Repeatedly square `x` T times mod `p`.
        for _ in 0..t {
            // y = y^2 mod p
            y = (&y * &y) % &self.p;
        }
        y
    }

    /// Decode a message `y` using the Sloth VDF.
    pub fn decode(y: &BigUint, p: &BigUint, t: u32) -> BigUint {
        let mut x = y.clone();
        for _ in 0..t {
            x = Sloth::mod_sqrt(&x, p);
        }
        x
    }

    /// Create a new tuple `(x, y)` where `x` is a random secret number
    /// and `y` is the encoded challenge value.
    #[cfg(not(target_arch = "wasm32"))]
    pub fn create(&self, t: u32) -> (BigUint, BigUint, BigUint) {
        let x = thread_rng().gen_biguint_range(&BigUint::one(), &self.p);
        let y = self.encode(&x, t);
        (x, y, self.p.clone())
    }

    /// Verify that the encoded value `y` was computed from the secret `x`.
    #[cfg(not(target_arch = "wasm32"))]
    pub fn verify(&self, x: &BigUint, y: &BigUint, t: u32) -> bool {
        let check = self.encode(x, t);
        &check == y
    }

    /// Compute a modular square root when p ≡ 3 mod 4:
    /// sqrt(a) mod p = a^((p+1)/4) mod p
    /// We'll pick the "lower root" consistently to ensure uniqueness.
    fn mod_sqrt(a: &BigUint, p: &BigUint) -> BigUint {
        // exponent = (p + 1) / 4
        let exp = (p + BigUint::one()) / 4_u32.to_biguint().unwrap();
        let root = a.modpow(&exp, p);

        // A prime p ≡ 3 mod 4 has exactly two roots: `r` and `p-r`.
        // We'll choose the smaller one to ensure uniqueness.
        let p_minus_root = (p - &root) % p;
        if root <= p_minus_root {
            root
        } else {
            p_minus_root
        }
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    #[test]
    fn test_sloth() {
        let p = crate::prime::generate_prime_mod_3_4(2048, 64);
        let t = 100;
        let sloth = Sloth::new(p);

        let (x, y, p) = sloth.create(t);
        assert!(sloth.verify(&x, &y, t));

        let decoded = Sloth::decode(&y, &p, t);
        assert!(sloth.verify(&decoded, &y, t));
    }

    #[test]
    fn test_sloth_incorrect() {
        let p = crate::prime::generate_prime_mod_3_4(2048, 64);
        let t = 100;
        let sloth = Sloth::new(p);

        let (x, y, _p) = sloth.create(t);
        assert!(sloth.verify(&x, &y, t));

        let decoded = 1024_u32.to_biguint().unwrap();
        assert!(!sloth.verify(&decoded, &y, t));
    }
}
